/**
 * file:    Subject.java
 *
 * classes: class Subject
 *
 * purpose: - implementation of a generic 'Subject'-server
 *
 * date:    05/03/97  last change: 05/03/97 18:00
 *
 * author:  Normen Seemann
 *
 * version: 1.00
 *
 */

package observer.generic;

import java.rmi.*;
import java.rmi.server.UnicastRemoteObject;
import java.util.*;


public abstract class Subject extends UnicastRemoteObject implements SubjectInterface {

    public static final int OBSERVERLOGGEDIN  = 0x0001;
    public static final int OBSERVERLOGGEDOUT = 0x0002;
    
    protected Critical observersCritical = new Critical(1);
    
    public Hashtable observers = new Hashtable();

    public Subject() throws RemoteException {

        super();

    }
    
    /*
     * login - method for an observer to login into a subject (considered as server)
     */
    public void login(ObserverInterface observer) throws RemoteException, AlreadyLoggedInException {

        String account;

        account = observer.getAccount();

        observersCritical.enter();
                    
        /*
         * ensure that there is no client named 'account' logged in yet
         */
        if(observers.containsKey(account)) {
            observersCritical.leave();
            System.out.println("throwing AlreadyLoggedInException");
            throw new AlreadyLoggedInException("Client is already logged in.");
        }

        /*
         * insert the observer into the 'observers' hashtable
         */
        observers.put(account, newObserverStorage(account, observer));
        
        System.out.println("Observer " + account + " has been registered.");
        
        broadcast(OBSERVERLOGGEDIN, (Object)account);
        
        observersCritical.leave();

    }

    /*
     * logout - method for an observer to logout
     */
    public void logout(ObserverInterface observer) throws RemoteException, NotLoggedInException {

        String account;

        account = observer.getAccount();

        observersCritical.enter();
        
        if((account == null) || (!observers.containsKey(account))) {
            observersCritical.leave();
            System.out.println("throwing NotLoggedInException");
            throw new NotLoggedInException("Client is not logged in.");
        }

        observers.remove(account);
        
        broadcast(OBSERVERLOGGEDOUT, (Object)account);
        
        System.out.println("Observer " + account + " has been removed.");
        
        observersCritical.leave();
        
    }

    /*
     * countObservers - count the registered observers
     */
    public int countObservers() throws RemoteException {

        return observers.size();

    }
    
    public String getInformation() throws RemoteException {
        
        return new String("This is a generic subject.");
        
    }
    
    /*
     * getObserverAccounts - returns an array of strings of accounts of all observers currently logged in 
     */
    public Vector getObserverAccounts() throws RemoteException {

        Vector o;
        Enumeration enumeration;
        int i;
        
        observersCritical.enter();
               
        o = new Vector(observers.size());
        enumeration = observers.keys();
        
        for(i = 0; i < observers.size(); i++)
            o.addElement(new String((String)enumeration.nextElement()));
            
        observersCritical.leave();
        
        return o;
    }

    /*
     * notifyChanges - method to notify changes in observers to their subject (normally initiates an observer update)
     */
    public void notifyChanges(ObserverInterface observer, int aspect) throws RemoteException {
        
        observersCritical.enter();
        broadcast(aspect, updateSubject(observer, aspect));
        observersCritical.leave();

    }

    protected abstract Object updateSubject(ObserverInterface observer, int aspect);

    protected void updateObserver(ObserverInterface observer, int aspect, Object param) throws RemoteException {

        System.out.println("Updating observer.");
        
        switch(aspect) {
            case OBSERVERLOGGEDIN:
                observer.observerLoggedIn((String)param);
                break;
                
            case OBSERVERLOGGEDOUT:
                observer.observerLoggedOut((String)param);
                break;

            default:
                observer.update(this, aspect);
            
        }
        
    }
    
    /*
     * broadcast - method for a subject to update its observers
     */
    protected void broadcast(int aspect, Object param) {

        Enumeration enumeration;
        String account;
        ObserverInterface observer;
        
        enumeration = observers.keys();

        while(enumeration.hasMoreElements()) {
            account = (String)enumeration.nextElement();

            try {
                observer = ((ObserverStorage)observers.get(account)).getObserver();
                updateObserver(observer, aspect, param);
            }
            catch(RemoteException e) {
                System.out.println("Subject.broadcast: an exception occurred: " + e.getMessage());
                e.printStackTrace();
            }
        }

    }
    
    protected ObserverStorage newObserverStorage(String a, ObserverInterface o) {

        return new ObserverStorage(a, o);
    
    }

}
