/**
 * file:    ApplicationClient.java
 *
 * classes: class ApplicationClient
 *
 * purpose: - a generic application specific client
 *
 * date:    05/09/97  last change: 05/03/97 18:00
 *
 * author:  Normen Seemann
 *
 * version: 1.00
 *
 */


package observer.main;

import java.rmi.*;
import java.rmi.server.UnicastRemoteObject;
import java.net.*;
import java.util.*;

import java.awt.*;
import java.awt.event.*;


import observer.generic.*;

/**
 * class ApplicationClient
 *
 */
public class ApplicationClient extends UnicastRemoteObject implements ApplicationClientInterface, ActionListener, Runnable {

    public static final int NOTACCEPTEDYET  = 0x0001;
    public static final int NOTACCEPTED     = 0x0002;
    public static final int ACCEPTED        = 0x0003;
    
    protected MainClientInterface mainClient;
    protected ApplicationServerInterface server;
    protected URL documentBase;
    protected Thread applicationThread = null;
    protected CenteredFrame sessionRequestedFrame = null;
    protected CenteredFrame waitingFrame = null;
    
    protected boolean isDead  = false;
    protected int accepted = NOTACCEPTEDYET;
    protected boolean clientsHaveAccepted = false;
    
    public ApplicationClient() throws RemoteException {
    
        super();
    
    }
    
    /*
     * returns the unique account of the observer (usually a username)
     */
    public String getAccount() throws RemoteException {

        return mainClient.getAccount();
    
    }

    /*
     * returns the current aspect of the observer (for example to indicate different observer-roles)
     */
    public int getAspect() throws RemoteException {
    
        return 0;
    
    }

    /*
     * method is called when a user has logged in
     */
    public void observerLoggedIn(String account) throws RemoteException {
    }

    /*
     * method is called when a user has logged out
     */
    public void observerLoggedOut(String account) throws RemoteException {
    }
    
    /*
     * updates the state of the observer
     */
    public void update(SubjectInterface si, int aspect) throws RemoteException {
    }
    
    /*
     * init - method as a replacement for a constructor. This is necessary because we need a no-arg constructor
     */
    public void init(MainClientInterface mC, ApplicationServerInterface aS, URL dB) throws RemoteException {
        
        mainClient = mC;
        server = aS;

        documentBase = dB;
    }
        
    /*
     * start - to start the application client
     */
    public void start() throws RemoteException{

        applicationThread = new Thread(this);
        applicationThread.start();
    
    }
    
    /*
     * destroy - to destroy the application client
     */
    public void destroy() throws RemoteException {
        
        synchronized(this) {
            isDead = true;
            notifyAll();
        }
        
        if(accepted != NOTACCEPTED) {
            try {
                applicationThread.join();
                applicationThread = null;
            }
            catch(Exception e)
            {
                System.out.println("ApplicationClient.destroy: an exception occurred: " +
    		                        e.getMessage());
        	    e.printStackTrace();
            }
        }
        
        if(sessionRequestedFrame != null) {
            sessionRequestedFrame.dispose();
            sessionRequestedFrame = null;
        }
        
        if(waitingFrame != null) {
            waitingFrame.dispose();
            waitingFrame = null;
        }
            
        mainClient = null;
        server = null;
    }
    
    public boolean isDead() throws RemoteException {
    
        return isDead;
    
    }
    
    public synchronized void clientsHaveAccepted() throws RemoteException {
       
        clientsHaveAccepted = true;
       
        notifyAll();
    
    }
    
    /*
     * getMainClient() - returns specific information about this subject
     */
    public MainClientInterface getMainClient() throws RemoteException {

        return mainClient;
    
    }
    
    /*
     * getServer - returns the corresponding server to this client
     */
    public ApplicationServerInterface getServer() throws RemoteException {

        return server;
    
    }

    /*
     * message - method to recieve a textual message
     */
    public void textMessage(String message) throws RemoteException {

        MessageBox messageBox = new MessageBox("Message from Server", message);
        messageBox.executeFrame(null);
    
    }

    /*
     * getDocumentBase - returns the document base of the applet 
     */
    public URL getDocumentBase() throws RemoteException {

        return documentBase;

    }
    
    public synchronized void actionPerformed(ActionEvent actionEvent) {

        Object source = actionEvent.getSource();
        
        if((sessionRequestedFrame != null) && (source == sessionRequestedFrame)) {
            
            if(((SimpleQuestionFrame)sessionRequestedFrame).getSelectedButton() == 0)
            {
                sessionRequestedFrame = null;
                accepted = ACCEPTED;                
            }
            else {
                sessionRequestedFrame = null;
                accepted = NOTACCEPTED;                
            }
            
            notifyAll();
        
        }
        
        if((waitingFrame != null) && (source == waitingFrame)) {
                waitingFrame = null;                
                accepted = NOTACCEPTED;
                notifyAll();        
        }


    }


    public void run() {

        if(!negotiateSession())
            return;
        
        // everything is fine - client can now start its actual work
            
    }
   
    protected void initRequestFrame() {

        Vector buttons = new Vector();
       
        buttons.addElement(new String("Yes"));
        buttons.addElement(new String("No"));
        sessionRequestedFrame = new SimpleQuestionFrame(this, "Session requested...", "The following persons would like to start a new session. Do you accept?", true, buttons);

        try {
            ((SimpleQuestionFrame)sessionRequestedFrame).appendItems(server.getObserverAccounts());
        }
        catch(Exception e) {
             System.out.println("ApplicationClient.run: an exception occurred: " + "Could not determine participants." + 
                                e.getMessage());
    	     e.printStackTrace();
        }

        sessionRequestedFrame = (SimpleQuestionFrame)sessionRequestedFrame.executeFrame(null);
    
    }
    
    protected void initWaitingFrame() {
        Vector buttons = new Vector();
        
        buttons.addElement(new String("Cancel"));
        waitingFrame = new SimpleQuestionFrame(this, "Session accepted...", "Waiting for other clients to accept...", false, buttons);
        waitingFrame = (SimpleQuestionFrame)waitingFrame.executeFrame(null);
    }
    
    protected boolean negotiateSession() {

        accepted = NOTACCEPTEDYET;
        
        clientsHaveAccepted = false;
        
        initRequestFrame();
        
        synchronized(this) {
            
            try {
                System.out.println("Waiting1");
                while(!isDead && (sessionRequestedFrame != null))
                    wait();
                System.out.println("Got It1");
                    
                if(sessionRequestedFrame != null) {
                    sessionRequestedFrame.dispose();
                    sessionRequestedFrame = null;
                }

                if(!isDead) {
                    if(accepted == ACCEPTED)
                        server.accept(this);
                    else {
                        if(accepted == NOTACCEPTED) {
                            mainClient.destroyClient(this, true);
                            return false;
                        }
                    }
                } else {
                    System.out.println("Returning.");
                    return false;
                }
            }
            catch(Exception e) {
                System.out.println("ApplicationClient.run: an exception occurred: " + "Could not determine participants." + 
                                   e.getMessage());
    	        e.printStackTrace();
            }

        }
        
        initWaitingFrame();
        
        synchronized(this) {
            
            try {
               
                System.out.println("Waiting2");
                while((!isDead) && (waitingFrame != null) && (!clientsHaveAccepted))
                    wait();
                    
                System.out.println("Got It2");
                                
                if(waitingFrame != null) {
                    waitingFrame.dispose();
                    waitingFrame = null;
                }
                
                if(!isDead && (accepted == NOTACCEPTED)) {
                    mainClient.destroyClient(this, true);
                    return false;
                }
                
                if(isDead)
                    return false;
                                        
            }
            catch(Exception e) {
                System.out.println("ApplicationClient.run: an exception occurred: " + "Could not determine participants." + 
                                   e.getMessage());
    	        e.printStackTrace();
            }

        }
        
        return true;

    }
    
    public void finalize() {

        System.out.println("ApplicationCLient destroyed.");
    
    }

}
