/**
 * file:    ApplicationServer.java
 *
 * classes: class ApplicationServer
 *
 * purpose: - an application specific server
 *
 * date:    05/09/97  last change: 05/03/97 18:00
 *
 * author:  Normen Seemann
 *
 * version: 1.00
 *
 */


package observer.main;

import java.rmi.*;
import java.util.*;

import observer.generic.*;

/**
 * class ApplicationServer
 *
 */
public class ApplicationServer extends Subject implements ApplicationServerInterface, Runnable {
 
    public static final int TEXTMESSAGE = 0x0101;
    
    protected MainServerInterface mainServer;
    protected String sessionName;
    
    protected Thread serverThread = null;
    protected boolean isDead = false;
    
    protected Vector toAccept = new Vector();
    protected boolean clientsHaveAccepted;
    
    public ApplicationServer() throws RemoteException {

        super();
        
    }
        
    /*
     * logout - method for an observer to logout
     */
    public void logout(ObserverInterface observer) throws RemoteException, NotLoggedInException {
    
        ApplicationClientInterface aC = (ApplicationClientInterface)observer;
        
        observersCritical.enter();
        
        super.logout(observer);
        
        observersCritical.leave();
                
        try {
            mainServer.destroyServer(getSessionName());
        }
        catch(NotRunningException e) {
            System.out.println("ApplicationServer.logout: an exception occurred: " +
                e.getMessage());
    	    e.printStackTrace();
        }
    
    }

    
    protected Object updateSubject(ObserverInterface observer, int aspect) {

        return null;

    }

    protected void updateObserver(ObserverInterface observer, int aspect, Object param) throws RemoteException {

        if(observer instanceof ApplicationClientInterface) {
            
            ApplicationClientInterface applicationClient = (ApplicationClientInterface)observer;
            
            switch(aspect) {
                case TEXTMESSAGE:
                    applicationClient.textMessage((String)param);
                    break;
                
                default:
                    super.updateObserver(observer, aspect, param);
            
            }
            
        } else
            super.updateObserver(observer, aspect, param);            
            
    }
    
    protected ObserverStorage newObserverStorage(String a, ObserverInterface o) {

        return new ApplicationClientStorage(a, o);
    
    }
        
    /*
     * init - method as a replacement for a constructor. This is necessary because we need a no-arg constructor
     */
    public void init(MainServerInterface mS, String sN, String cCN, Vector o) throws RemoteException {
        
        ApplicationClientInterface aC;
        
        mainServer = mS;
        sessionName = new String(sN);
        
        for(int i = 0; i < o.size(); i++)
            ((MainClientInterface)o.elementAt(i)).startClient(this, cCN);
            
            
    }
    
    public void start() throws RemoteException {
        
        serverThread = new Thread(this);
        serverThread.start();
        
    }
    
    public void destroy() throws RemoteException {

        Enumeration enum;
        ApplicationClientInterface aC;
        
        System.out.println("Destroying server...");
        
        synchronized(this) {
            if(isDead)
                return;
            
            isDead = true;
            notifyAll();
        }

        try {
            serverThread.join();
            serverThread = null;
        }
        catch(Exception e)
        {
            System.out.println("ApplicationServer.destroy: an exception occurred: " +
		                        e.getMessage());
    	    e.printStackTrace();
        }
        
        observersCritical.enter();
        
        enum = observers.elements();
        
        while(enum.hasMoreElements()) {
            aC = (ApplicationClientInterface)((ApplicationClientStorage)enum.nextElement()).getObserver();

            // this complicated form is necessary,
            // because on every client-machine there is a pool of active clients,
            // this pool must stay consistent
            aC.getMainClient().destroyClient(aC, false);
        }
            
        observers = null;
        
        observersCritical.leave();
        
        System.out.println("Server destroyed.");
    
    }
    
    public void accept(ApplicationClientInterface aC) throws RemoteException {
        
        String account = aC.getAccount(); 
        
        observersCritical.enter();
        
        System.out.println("Client " + account + " has accepted.");
        
        toAccept.removeElement(account);
        
        ((ApplicationClientStorage)observers.get(account)).accept();
        
        if(toAccept.isEmpty()) {
            
            System.out.println("All clients have accepted.");

            synchronized(this) {
                clientsHaveAccepted = true;
                notifyAll();
            }

        }
        
        observersCritical.leave();
    }
    
    /*
     * getSessionName() - returns the session name of this session
     */
    public String getSessionName() throws RemoteException {
    
        return sessionName;
    
    }
    
    /*
     * getInformation() - returns specific information about this subject
     */
    public String getInformation() throws RemoteException {
        
        return new String("This is a generic application server.");
        
    }

    public void run() {
        
        negotiateSession();

    }
        
    protected void negotiateSession() {
    
        Enumeration enum;
        ApplicationClientInterface aC;
        
        clientsHaveAccepted = false;

        observersCritical.enter();
        
        enum = observers.elements();
        
        while(enum.hasMoreElements()) {
            aC = (ApplicationClientInterface)((ApplicationClientStorage)enum.nextElement()).getObserver();

            try {
                // start clients
                toAccept.addElement(aC.getAccount());
                informClient(aC);
            }
            catch(Exception e) {
                System.out.println("ApplicationServer.run: an exception occurred: " +
	    	                        e.getMessage());
    	        e.printStackTrace();
            }
        }
        
        observersCritical.leave();
        
        synchronized(this) {
            
            try {
                while(!isDead && !clientsHaveAccepted)
                    wait();
            }
            catch(Exception e) {
                System.out.println("ApplicationServer.negotiate: an exception occurred: " +
	    	                        e.getMessage());
    	        e.printStackTrace();
            }
                
            if(isDead)
                return;
        
        }
        
        System.out.println("Clients have accepted. Going to inform them.");
        
        observersCritical.enter();
        
        enum = observers.elements();
        
        while(enum.hasMoreElements()) {
            aC = (ApplicationClientInterface)((ApplicationClientStorage)enum.nextElement()).getObserver();

            try {
                // clients have to be accepted
                aC.clientsHaveAccepted();
            }
            catch(Exception e) {
                System.out.println("ApplicationServer.run: an exception occurred: " +
	    	                        e.getMessage());
    	        e.printStackTrace();
            }
        }
        
        observersCritical.leave();
        
        // everything is fine - the server can now start its actual work    
    }
    

    // this method is called when negotiateSession() needs to inform the clients
    protected void informClient(ApplicationClientInterface aC) throws RemoteException {

        aC.start();
        
    }

}
