/**
 * file:    MainServer.java
 *
 * classes: class MainServer
 *
 * purpose: - a main server is used to lauch other 'Subject'-server
 *          - it's supposed to be the only one which uses the registry to
 *            bind itself to a URL
 *
 * date:    05/03/97  last change: 05/03/97 18:00
 *
 * author:  Normen Seemann
 *
 * version: 1.00
 *
 */

package observer.main;

import java.rmi.*;
import java.rmi.registry.LocateRegistry;
import java.rmi.server.UnicastRemoteObject;
import java.util.*;

import observer.generic.*;

public class MainServer extends Subject implements MainServerInterface {
    
    public static final int SESSIONCREATED   = 0x0010;
    public static final int SESSIONDESTROYED = 0x0011;
    
    private Vector viewers = new Vector();

    private Critical runningServersCritical = new Critical(1);
    
    private Hashtable installedSystems = new Hashtable();
    private Hashtable runningServers = new Hashtable();        

    public MainServer() throws RemoteException {
        super();
    }
    
    /*
     * login - overrides method in Subject and adds functionality for managing viewers
     */
    public void login(ObserverInterface observer) throws RemoteException, AlreadyLoggedInException {
        
        if(!(observer instanceof MainClientInterface))
            throw new RemoteException("Observer is not compatible.");
            
        super.login(observer);
            
    }
    
    /*
     * logout - overrides method in Subject and adds functionality for managing viewers
     */
    public void logout(ObserverInterface observer) throws RemoteException, NotLoggedInException {
        
        try {
          super.logout(observer);
        }
        catch(NotLoggedInException e) {

            System.out.println("A viewer has been removed.");
            if(!viewers.removeElement(observer))
                throw e;
        }
        
    }

    protected Object updateSubject(ObserverInterface observer, int aspect) {

        return null;

    }
    
    protected void updateObserver(ObserverInterface observer, int aspect, Object param) throws RemoteException {

        if(observer instanceof MainClientInterface) {
            
            MainClientInterface mainClient = (MainClientInterface)observer;
            
            switch(aspect) {
                case SESSIONCREATED:
                    mainClient.sessionCreated((String)param);
                    break;
                
                case SESSIONDESTROYED:
                    mainClient.sessionDestroyed((String)param);
                    break;

                default:
                    super.updateObserver(observer, aspect, param);
            
            }
            
        } else
            super.updateObserver(observer, aspect, param);            
            
    }

    /*
     * broadcast - method for the mainserver to update its observers and viewers
     */
    protected void broadcast(int aspect, Object param) {

        ObserverInterface viewer;

        /*
         * first of all, call the super.broadcast method
         */
        super.broadcast(aspect, param);

        for(int i = 0; i < viewers.size(); i ++) {
            viewer = (ObserverInterface)viewers.elementAt(i);

            try {
                updateObserver(viewer, aspect, param);
            }
            catch(RemoteException e) {
                System.out.println("MainServer.broadcast: an exception occurred: " + e.getMessage());
                e.printStackTrace();
            }
        }

    }

    /*
     * registerViewer - method for an observer to login into a subject (considered as server)
     * Note: Since we don't have a safe login mechanism as in
     *       login (see SubjectInterface) a viewer can register several times.
     *       However this is not recommended, because viewers registered more than once
     *       will receive multiple update commands
     */
    public void registerViewer(ObserverInterface observer) throws RemoteException {

        if(!(observer instanceof MainClientInterface))
            throw new RemoteException("Observer is not compatible.");
                
        viewers.addElement(observer);

        System.out.println("A viewer has been registered.");

    }
    
    /*
     * getRunningServers() returns the names of all running servers
     */
    public Vector getRunningServers() throws RemoteException {
        
        Vector result;
        Enumeration e;
        
        runningServersCritical.enter();
        
        result = new Vector(runningServers.size());
        e = runningServers.keys();
        
        while(e.hasMoreElements())
            result.addElement(new String((String)e.nextElement()));
            
        runningServersCritical.leave();
                
        return result;
    
    }
    
    /*
     * getServerInformation(String serverName) returns information about thge specified server
     */
    public String getServerInformation(String serverName)
        throws RemoteException, NotRunningException {
            
            ApplicationServerInterface server;
            String result;
            
            runningServersCritical.enter();
            
            if(!runningServers.containsKey(serverName)) {
                runningServersCritical.leave();
                throw new NotRunningException("Server " + serverName + " is not running currently.");
            }
            
            server = (ApplicationServerInterface)runningServers.get(serverName);
            
            result = new String(server.getInformation());
                        
            runningServersCritical.leave();
            
            return result;
            
    }
    
    /*
     * getInstalledSystems() returns the names of systems currently installed on the server
     */
    public Vector getInstalledSystems() throws RemoteException {

        Vector result;
        Enumeration e;
        
        result = new Vector(installedSystems.size());
        e = installedSystems.keys();
        
        while(e.hasMoreElements())
            result.addElement(new String((String)e.nextElement()));
                
        return result;
    
    }
    
    /*
     * startServer() tries to start a server of the specified system using an array
     * of accounts to determine the observers that will participate
     */
    public void startServer(String systemString, String sessionName, Vector clientAccounts)
        throws RemoteException, NotLoggedInException, NotInstalledException, AlreadyRunningException, UserMismatchException {

        Vector v;
        Vector observerVector;
        Class serverClass;
        ApplicationServerInterface server;
        
        System.out.println("in StartServer");

        runningServersCritical.enter();        
        observersCritical.enter();

        observerVector = new Vector();
        
        for(int i = 0; i < clientAccounts.size(); i ++) {
            Object observer = ((ObserverStorage)observers.get((String)clientAccounts.elementAt(i))).getObserver();
            if(observer != null) {
                System.out.println((String)clientAccounts.elementAt(i) + " resolved.");
                observerVector.addElement(observer);
            } else {
                observersCritical.leave();
                runningServersCritical.leave();
                throw new NotLoggedInException("Client is not logged in.");
            }
        }
        
        if((v = (Vector)installedSystems.get(systemString)) == null) {
            observersCritical.leave();
            runningServersCritical.leave();
            throw new NotInstalledException("Could not find the server " + systemString + ".");
        }
        
        if(runningServers.containsKey(sessionName)) {
            observersCritical.leave();
            runningServersCritical.leave();
            throw new AlreadyRunningException("Session " + sessionName + " is already running.");
        }
        
        if((observerVector.size() < ((Integer)v.elementAt(2)).intValue()) ||
           (observerVector.size() > ((Integer)v.elementAt(3)).intValue())) {

            observersCritical.leave();
            runningServersCritical.leave();
            throw new UserMismatchException(((Integer)v.elementAt(2)).intValue(), ((Integer)v.elementAt(3)).intValue());
        
        }
                           
        try {
            serverClass = Class.forName((String)v.elementAt(0));
            runningServers.put(sessionName, server = (ApplicationServerInterface)serverClass.newInstance());
        }
        catch(Exception e) {
  	        System.out.println("MainServer.startServer: an exception occurred: " +
                e.getMessage());
    	    e.printStackTrace();
            observersCritical.leave();
            runningServersCritical.leave();
    	    throw new RemoteException("Could not start server:", e);
        }
        
        server.init(this, sessionName, (String)v.elementAt(1), observerVector);
        
        server.start();
        
        observersCritical.leave();
        
        broadcast(SESSIONCREATED, sessionName);
        
        runningServersCritical.leave();

    }
    
    public void destroyServer(String sessionName) throws RemoteException, NotRunningException {

        ApplicationServerInterface server;
        
        runningServersCritical.enter();
        
        server = (ApplicationServerInterface)runningServers.get(sessionName);
        
        server.destroy();

        runningServers.remove(sessionName);
        
        broadcast(SESSIONDESTROYED, sessionName);
        
        runningServersCritical.leave();
    
    }

    public void installSystem(String systemName, String serverClassName, String clientClassName, int min, int max) {
        
        Vector v = new Vector(4);
        
        v.addElement(serverClassName);
        v.addElement(clientClassName);
        v.addElement(new Integer(min));
        v.addElement(new Integer(max));
        
        installedSystems.put(systemName, v);
        
        System.out.println("System " + systemName + " has been inserted successfully.");
        
    }
        
    /**
     * Start up the MainServer; also creates a registry so that the
     * Observer can lookup the server.
     */
    public static void main(String args[])
    {
        MainServer mainServer;
    	
    	// Create and install the security manager
	    System.setSecurityManager(new RMISecurityManager());
    	
    	try {
	        System.out.println("MainServer.main: creating registry");
	        LocateRegistry.createRegistry(2005);

    	    System.out.println("MainServer.main: creating server");
	        mainServer = new MainServer();

	        System.out.println("MainServer.main: binding server ");
    	    Naming.rebind("//:2005/observer.main.MainServer", mainServer);
	        System.out.println("MainServer.main: done");
    	} catch (Exception e) {
	        System.out.println("MainServer.main: an exception occurred: " +
		    	       e.getMessage());
    	    e.printStackTrace();
    	    
    	    return;
	    }
	    
	    mainServer.installSystem("generic", "observer.main.ApplicationServer", "observer.main.ApplicationClient", 1, 0);
	    mainServer.installSystem("Gomuko", "observer.gomuko.GomukoServer", "observer.gomuko.GomukoClient", 2, 2);
	    
    }

}
